local AddonName, Addon = ...

-- Lib
local DGL = Addon.Libs.DGL

-- Upvalues
local abs = math.abs

local Clamp, PlaySoundFile = Clamp, PlaySoundFile

-- Game
local Player = Addon.Player
local Game = Addon.Game
local Spritesheet = Addon.Spritesheet
local World = Addon.World

-- Mixins
local Mixins = {}

-- Animations
local ANIMATIONS = {
  Path = Spritesheet.Path,
  Run = {
    FrameRate = 1 / 10, -- 10 fps
    Frames = {
      Spritesheet.Sprites.BirdDown.Coords,
      Spritesheet.Sprites.BirdUp.Coords
    }
  },
  Die = {
    FrameRate = 0,
    Frames = {
      Spritesheet.Sprites.BirdDead.Coords
    }
  }
}

-- Consts
local DIE_SOUND_PATH = format("Interface\\Addons\\%s\\Assets\\bump.ogg", AddonName)

local PI, HALF_PI, QUARTER_PI = math.pi, math.pi * 0.5, math.pi * 0.25

local WIDTH = Spritesheet.Sprites.BirdUp.Width
local HEIGHT = Spritesheet.Sprites.BirdUp.Height
local OFFSET_X = WIDTH * 0.25
local OFFSET_Y = HEIGHT * 0.25
local START_X = Game.HALF_WIDTH - OFFSET_X
local START_Y = Game.HALF_HEIGHT - OFFSET_Y
local GROUND_HEIGHT = Game.HEIGHT - Spritesheet.Sprites.Ground.Height - OFFSET_Y

local JUMP_DELAY = 0.05
local JUMP_VELOCITY = 300
local GRAVITY = 10

-- Variables
local velocity = 0
local jumpDelay = 0

-- ============================================================================
-- General Functions
-- ============================================================================

function Player:Initialize()
  local player = DGL.GameObject:Create(Game, "Player", START_X, START_Y, "ARTWORK", Addon.Sublayers.Player)
  player:SetWidth(WIDTH)
  player:SetHeight(HEIGHT)
  player:SetOffset(OFFSET_X, OFFSET_Y)

  World:add(player, player.X, player.Y, player.HALF_WIDTH, player.HALF_HEIGHT)

  player:SetAnimations(ANIMATIONS)
  player:PlayAnimation("Run")

  -- Add mixins
  for k, v in next, Mixins do player[k] = v end

  --[[ Debug collision box
  do
    local t = Game.WorldFrame:CreateTexture(nil, "ARTWORK", nil, Addon.Sublayers.Player+1)
    t:SetSize(player.HALF_WIDTH, player.HALF_HEIGHT)
    t:SetColorTexture(1, 0, 1, 1)
    t:SetPoint("TOPLEFT", player.Texture, OFFSET_X, -OFFSET_Y)
  end
  --]]
  
  self.player = player
  self.Initialize = nil
end

function Player:Die()
  PlaySoundFile(DIE_SOUND_PATH)
  self.player:PlayAnimation("Die", true)
  self.player.isAlive = false
  velocity = abs(velocity)
end

function Player:Reset()
  local player = self.player
  velocity = 0
  jumpDelay = JUMP_DELAY

  player.X = START_X
  player.Y = START_Y
  player.isAlive = true

  World:update(player, player.X, player.Y)
  
  player.Texture:SetRotation(0)
  player:PlayAnimation("Run")
end

-- ============================================================================
-- Player GameObject Mixins
-- ============================================================================

function Mixins:OnUpdate(elapsed)
  velocity = velocity - GRAVITY

  if self.isAlive then
    -- Update rotation
    self.Texture:SetRotation(
      max(self.Texture:GetRotation() - 0.025, -HALF_PI)
    )

    -- Update jump state
    jumpDelay = jumpDelay + elapsed
    if
      (jumpDelay >= JUMP_DELAY) and
      (
        Game.Input:IsMouseJustDown("LeftButton") or
        Game.Input:IsKeyJustDown("SPACE")
      )
    then
      jumpDelay = 0
      velocity = JUMP_VELOCITY
      self.Texture:SetRotation(QUARTER_PI)
    end

    -- Update position
    local cols, cols_len
    self.X, self.Y, cols, cols_len = World:move(self, self.X, self.Y - (velocity * elapsed))
    
    -- Collision
    for i=1, cols_len do
      local col = cols[i]
      if (col.other.Name == "Column") then Game:GameOver() end
    end
  else
    -- Update rotation
    self.Texture:SetRotation(max(self.Texture:GetRotation() - 0.05, -PI))
    -- Update position
    self.Y = self.Y - (velocity * elapsed)
  end

  -- Clamp y, this can cause the player's world rect y value to differ, but it
  -- clears back up immediately on the next call to move(). So, a call to
  -- World:update() should not be necessary.
  self.Y = Clamp(self.Y, OFFSET_Y, GROUND_HEIGHT)
  -- World:update(self, self.X, self.Y)

  -- If touching top, immediately begin falling
  if (self.Y <= OFFSET_Y) then velocity = 0 end
  -- If touching ground, GameOver!
  if (self.Y >= GROUND_HEIGHT) then
    if self.isAlive then
      Game:GameOver()
    else
      Game:SetState(Game.States.StartGame)
      self.Texture:SetRotation(-PI)
    end
  end
end
